@@ -12,7 +12,7 @@ from collections import defaultdict
 import logging
 import hashlib
 from pathlib import Path
-from typing import List, Optional, Union, IO
+from typing import List, Optional, Union, IO, Dict, Any
 
 from .abstract import AbstractMISP, MISPTag
 from .exceptions import UnknownMISPObjectTemplate, InvalidMISPObject, PyMISPError, NewEventError, NewAttributeError
@@ -82,7 +82,7 @@ def _make_datetime(value) -> datetime:
     return value
 
 
-def make_bool(value: Union[bool, int, str, dict, list, None]) -> bool:
+def make_bool(value: Optional[Union[bool, int, str, dict, list]]) -> bool:
     if isinstance(value, bool):
         return value
     if isinstance(value, int):
@@ -102,6 +102,10 @@ class MISPOrganisation(AbstractMISP):
 
     _fields_for_feed: set = {'name', 'uuid'}
 
+    def __init__(self):
+        super().__init__()
+        self.id: int
+
     def from_dict(self, **kwargs):
         if 'Organisation' in kwargs:
             kwargs = kwargs['Organisation']
@@ -167,19 +171,20 @@ class MISPSighting(AbstractMISP):
 
 class MISPAttribute(AbstractMISP):
     _fields_for_feed: set = {'uuid', 'value', 'category', 'type', 'comment', 'data',
-                             'timestamp', 'to_ids', 'disable_correlation', 'first_seen', 'last_seen'}
+                             'deleted', 'timestamp', 'to_ids', 'disable_correlation',
+                             'first_seen', 'last_seen'}
 
-    def __init__(self, describe_types: Optional[dict]=None, strict: bool=False):
+    def __init__(self, describe_types: Optional[Dict]=None, strict: bool=False):
         """Represents an Attribute
             :describe_type: Use it is you want to overwrite the defualt describeTypes.json file (you don't)
             :strict: If false, fallback to sane defaults for the attribute type if the ones passed by the user are incorrect
         """
         super().__init__()
         if describe_types:
-            self.describe_types: dict = describe_types
+            self.describe_types: Dict[str, Any] = describe_types
         self.__categories: List[str] = self.describe_types['categories']
-        self.__category_type_mapping: dict = self.describe_types['category_type_mappings']
-        self.__sane_default: dict = self.describe_types['sane_defaults']
+        self.__category_type_mapping: Dict[str, List[str]] = self.describe_types['category_type_mappings']
+        self.__sane_default: Dict[str, Dict[str, Union[str, int]]] = self.describe_types['sane_defaults']
         self.__strict: bool = strict
         self.data: Optional[BytesIO] = None
         self.first_seen: datetime
@@ -194,7 +199,7 @@ class MISPAttribute(AbstractMISP):
         self.Event: MISPEvent
         self.RelatedAttribute: List[MISPAttribute]
 
-    def add_tag(self, tag: Optional[Union[str, MISPTag, dict]]=None, **kwargs) -> MISPTag:
+    def add_tag(self, tag: Optional[Union[str, MISPTag, Dict]]=None, **kwargs) -> MISPTag:
         return super()._add_tag(tag, **kwargs)
 
     @property
@@ -207,7 +212,7 @@ class MISPAttribute(AbstractMISP):
         """Set a list of prepared MISPTag."""
         super()._set_tags(tags)
 
-    def _prepare_data(self, data: Union[Path, str, bytes, BytesIO, None]):
+    def _prepare_data(self, data: Optional[Union[Path, str, bytes, BytesIO]]):
         if not data:
             super().__setattr__('data', None)
             return
@@ -239,15 +244,15 @@ class MISPAttribute(AbstractMISP):
                 # not a encrypted zip file, assuming it is a new malware sample
                 self._prepare_new_malware_sample()
 
-    def __setattr__(self, name, value):
+    def __setattr__(self, name: str, value: Any):
         if name in ['first_seen', 'last_seen']:
-            value = _make_datetime(value)
+            _datetime = _make_datetime(value)
 
-            if name == 'last_seen' and hasattr(self, 'first_seen') and self.first_seen > value:
+            if name == 'last_seen' and hasattr(self, 'first_seen') and self.first_seen > _datetime:
                 raise PyMISPError('last_seen ({value}) has to be after first_seen ({self.first_seen})')
-            if name == 'first_seen' and hasattr(self, 'last_seen') and self.last_seen < value:
+            if name == 'first_seen' and hasattr(self, 'last_seen') and self.last_seen < _datetime:
                 raise PyMISPError('first_seen ({value}) has to be before last_seen ({self.last_seen})')
-            super().__setattr__(name, value)
+            super().__setattr__(name, _datetime)
         elif name == 'data':
             self._prepare_data(value)
         else:
@@ -278,7 +283,7 @@ class MISPAttribute(AbstractMISP):
         if not hasattr(self, 'timestamp'):
             self.timestamp = datetime.timestamp(datetime.now())
 
-    def _to_feed(self) -> dict:
+    def _to_feed(self) -> Dict:
         to_return = super()._to_feed()
         if self.data:
             to_return['data'] = base64.b64encode(self.data.getvalue()).decode()
@@ -292,7 +297,7 @@ class MISPAttribute(AbstractMISP):
         return self.describe_types['types']
 
     @property
-    def malware_binary(self) -> Union[BytesIO, None]:
+    def malware_binary(self) -> Optional[BytesIO]:
         """Returns a BytesIO of the malware (if the attribute has one, obvs)."""
         if hasattr(self, '_malware_binary'):
             return self._malware_binary
@@ -330,7 +335,7 @@ class MISPAttribute(AbstractMISP):
         """Alias for add_shadow_attribute"""
         return self.add_shadow_attribute(shadow_attribute, **kwargs)
 
-    def add_shadow_attribute(self, shadow_attribute: Union[MISPShadowAttribute, dict, None]=None, **kwargs) -> MISPShadowAttribute:
+    def add_shadow_attribute(self, shadow_attribute: Optional[Union[MISPShadowAttribute, Dict]]=None, **kwargs) -> MISPShadowAttribute:
         """Add a shadow attribute to the attribute (by name or a MISPShadowAttribute object)"""
         if isinstance(shadow_attribute, MISPShadowAttribute):
             misp_shadow_attribute = shadow_attribute
@@ -346,7 +351,7 @@ class MISPAttribute(AbstractMISP):
         self.edited = True
         return misp_shadow_attribute
 
-    def add_sighting(self, sighting: Union[MISPSighting, dict, None]=None, **kwargs) -> MISPSighting:
+    def add_sighting(self, sighting: Optional[Union[MISPSighting, dict]]=None, **kwargs) -> MISPSighting:
         """Add a sighting to the attribute (by name or a MISPSighting object)"""
         if isinstance(sighting, MISPSighting):
             misp_sighting = sighting
@@ -488,7 +493,7 @@ class MISPAttribute(AbstractMISP):
 
         super().from_dict(**kwargs)
 
-    def to_dict(self) -> dict:
+    def to_dict(self) -> Dict:
         to_return = super().to_dict()
         if self.data:
             to_return['data'] = base64.b64encode(self.data.getvalue()).decode()
@@ -595,7 +600,8 @@ class MISPObject(AbstractMISP):
 
     _fields_for_feed: set = {'name', 'meta-category', 'description', 'template_uuid',
                              'template_version', 'uuid', 'timestamp', 'distribution',
-                             'sharing_group_id', 'comment', 'first_seen', 'last_seen'}
+                             'sharing_group_id', 'comment', 'first_seen', 'last_seen',
+                             'deleted'}
 
     def __init__(self, name: str, strict: bool=False, standalone: bool=False, default_attributes_parameters: dict={}, **kwargs):
         ''' Master class representing a generic MISP object
@@ -623,7 +629,7 @@ class MISPObject(AbstractMISP):
         self.last_seen: datetime
         self.__fast_attribute_access: dict = defaultdict(list)  # Hashtable object_relation: [attributes]
         self.ObjectReference: List[MISPObjectReference] = []
-        self.Attribute: List[MISPAttribute] = []
+        self.Attribute: List[MISPObjectAttribute] = []
         self.SharingGroup: MISPSharingGroup
         self._default_attributes_parameters: dict
         if isinstance(default_attributes_parameters, MISPAttribute):
@@ -645,8 +651,8 @@ class MISPObject(AbstractMISP):
             self._default_attributes_parameters.pop('data', None)  # in case the original in a sample or an attachment
 
             # Those values are set for the current object, if they exist, but not pop'd because they are still useful for the attributes
-            self.distribution = self._default_attributes_parameters.get('distribution', 5)
-            self.sharing_group_id = self._default_attributes_parameters.get('sharing_group_id', 0)
+            self.distribution: int = self._default_attributes_parameters.get('distribution', 5)
+            self.sharing_group_id: int = self._default_attributes_parameters.get('sharing_group_id', 0)
         else:
             self.distribution = 5  # Default to inherit
             self.sharing_group_id = 0
@@ -656,7 +662,7 @@ class MISPObject(AbstractMISP):
             self.update_not_jsonable('ObjectReference')
 
     def _load_template_path(self, template_path: Union[Path, str]) -> bool:
-        self._definition: Union[dict, None] = self._load_json(template_path)
+        self._definition: Optional[Dict] = self._load_json(template_path)
         if not self._definition:
             return False
         setattr(self, 'meta-category', self._definition['meta-category'])
@@ -671,7 +677,7 @@ class MISPObject(AbstractMISP):
         if not hasattr(self, 'timestamp'):
             self.timestamp = datetime.timestamp(datetime.now())
 
-    def _to_feed(self) -> dict:
+    def _to_feed(self) -> Dict:
         to_return = super(MISPObject, self)._to_feed()
         if self.references:
             to_return['ObjectReference'] = [reference._to_feed() for reference in self.references]
@@ -714,11 +720,11 @@ class MISPObject(AbstractMISP):
         self._strict = False
 
     @property
-    def attributes(self) -> List[MISPAttribute]:
+    def attributes(self) -> List['MISPObjectAttribute']:
         return self.Attribute
 
     @attributes.setter
-    def attributes(self, attributes: List[MISPAttribute]):
+    def attributes(self, attributes: List['MISPObjectAttribute']):
         if all(isinstance(x, MISPObjectAttribute) for x in attributes):
             self.Attribute = attributes
             self.__fast_attribute_access = defaultdict(list)
@@ -826,24 +832,31 @@ class MISPObject(AbstractMISP):
         return self._fast_attribute_access.get(object_relation, [])
 
     @property
-    def _fast_attribute_access(self):
+    def _fast_attribute_access(self) -> Dict:
         if not self.__fast_attribute_access:
             for a in self.attributes:
                 self.__fast_attribute_access[a.object_relation].append(a)
         return self.__fast_attribute_access
 
-    def has_attributes_by_relation(self, list_of_relations: List[str]):
+    def has_attributes_by_relation(self, list_of_relations: List[str]) -> bool:
         '''True if all the relations in the list are defined in the object'''
         return all(relation in self._fast_attribute_access for relation in list_of_relations)
 
-    def add_attribute(self, object_relation: str, simple_value: Union[str, int, float]=None, **value) -> Union[MISPAttribute, None]:
+    def add_attribute(self, object_relation: str, simple_value: Optional[Union[str, int, float]]=None, **value) -> Optional[MISPAttribute]:
         """Add an attribute. object_relation is required and the value key is a
         dictionary with all the keys supported by MISPAttribute"""
         if simple_value is not None:  # /!\ The value *can* be 0
             value = {'value': simple_value}
-        if value.get('value') in [None, '']:
-            logger.warning("The value of the attribute you're trying to add is None or empty string, skipping it. Object relation: {}".format(object_relation))
+        if value.get('value') is None:
+            logger.warning("The value of the attribute you're trying to add is None, skipping it. Object relation: {}".format(object_relation))
             return None
+        else:
+            # Make sure we're not adding an empty value.
+            if isinstance(value['value'], str):
+                value['value'] = value['value'].strip()
+                if value['value'] == '':
+                    logger.warning("The value of the attribute you're trying to add is an empty string, skipping it. Object relation: {}".format(object_relation))
+                    return None
         if self._known_template and self._definition:
             if object_relation in self._definition['attributes']:
                 attribute = MISPObjectAttribute(self._definition['attributes'][object_relation])
@@ -876,7 +889,7 @@ class MISPObject(AbstractMISP):
             to_return.append(a)
         return to_return
 
-    def to_dict(self, strict: bool=False) -> dict:
+    def to_dict(self, strict: bool=False) -> Dict:
         if strict or self._strict and self._known_template:
             self._validate()
         return super(MISPObject, self).to_dict()
@@ -886,10 +899,12 @@ class MISPObject(AbstractMISP):
             self._validate()
         return super(MISPObject, self).to_json(sort_keys=sort_keys, indent=indent)
 
-    def _validate(self):
+    def _validate(self) -> bool:
+        if not self._definition:
+            raise PyMISPError('No object definition available, unable to validate.')
         """Make sure the object we're creating has the required fields"""
         if self._definition.get('required'):
-            required_missing = set(self._definition.get('required')) - set(self._fast_attribute_access.keys())
+            required_missing = set(self._definition['required']) - set(self._fast_attribute_access.keys())
             if required_missing:
                 raise InvalidMISPObject('{} are required.'.format(required_missing))
         if self._definition.get('requiredOneOf'):
@@ -916,7 +931,7 @@ class MISPEvent(AbstractMISP):
     _fields_for_feed: set = {'uuid', 'info', 'threat_level_id', 'analysis', 'timestamp',
                              'publish_timestamp', 'published', 'date', 'extends_uuid'}
 
-    def __init__(self, describe_types: dict=None, strict_validation: bool=False, **kwargs):
+    def __init__(self, describe_types: Optional[Dict]=None, strict_validation: bool=False, **kwargs):
         super().__init__(**kwargs)
         if strict_validation:
             schema_file = 'schema.json'
@@ -971,7 +986,7 @@ class MISPEvent(AbstractMISP):
             self.threat_level_id = 4
 
     @property
-    def manifest(self) -> dict:
+    def manifest(self) -> Dict:
         required = ['info', 'Orgc']
         for r in required:
             if not hasattr(self, r):
@@ -1000,7 +1015,7 @@ class MISPEvent(AbstractMISP):
                 to_return += attribute.hash_values(algorithm)
         return to_return
 
-    def to_feed(self, valid_distributions: List[int]=[0, 1, 2, 3, 4, 5], with_meta: bool=False) -> dict:
+    def to_feed(self, valid_distributions: List[int]=[0, 1, 2, 3, 4, 5], with_meta: bool=False) -> Dict:
         """ Generate a json output for MISP Feed.
         Notes:
             * valid_distributions only makes sense if the distribution key is set (i.e. the event is exported from a MISP instance)
@@ -1090,7 +1105,7 @@ class MISPEvent(AbstractMISP):
             raise PyMISPError('All the attributes have to be of type MISPShadowAttribute.')
 
     @property
-    def related_events(self):  # -> List[MISPEvent]:
+    def related_events(self) -> List['MISPEvent']:
         return self.RelatedEvent
 
     @property
@@ -1233,7 +1248,7 @@ class MISPEvent(AbstractMISP):
             self.SharingGroup.from_dict(**kwargs.pop('SharingGroup'))
         super(MISPEvent, self).from_dict(**kwargs)
 
-    def to_dict(self) -> dict:
+    def to_dict(self) -> Dict:
         to_return = super().to_dict()
 
         if to_return.get('date'):
@@ -1276,7 +1291,7 @@ class MISPEvent(AbstractMISP):
             if ((hasattr(a, 'id') and a.id == attribute_identifier)
                     or (hasattr(a, 'uuid') and a.uuid == attribute_identifier)
                     or (hasattr(a, 'value') and attribute_identifier == a.value
-                        or attribute_identifier in a.value.split('|'))):
+                        or (isinstance(a.value, str) and attribute_identifier in a.value.split('|')))):
                 tags += a.tags
         return tags
 
@@ -1290,7 +1305,7 @@ class MISPEvent(AbstractMISP):
             if ((hasattr(a, 'id') and a.id == attribute_identifier)
                     or (hasattr(a, 'uuid') and a.uuid == attribute_identifier)
                     or (hasattr(a, 'value') and attribute_identifier == a.value
-                        or attribute_identifier in a.value.split('|'))):
+                        or (isinstance(a.value, str) and attribute_identifier in a.value.split('|')))):
                 a.add_tag(tag)
                 attributes.append(a)
 
@@ -1309,14 +1324,12 @@ class MISPEvent(AbstractMISP):
 
     def delete_attribute(self, attribute_id: str):
         """Delete an attribute, you can search by ID or UUID"""
-        found = False
         for a in self.attributes:
             if ((hasattr(a, 'id') and a.id == attribute_id)
                     or (hasattr(a, 'uuid') and a.uuid == attribute_id)):
                 a.delete()
-                found = True
                 break
-        if not found:
+        else:
             raise PyMISPError('No attribute with UUID/ID {} found.'.format(attribute_id))
 
     def add_attribute(self, type: str, value: Union[str, int, float], **kwargs) -> Union[MISPAttribute, List[MISPAttribute]]:
@@ -1501,6 +1514,8 @@ class MISPFeed(AbstractMISP):
         if 'Feed' in kwargs:
             kwargs = kwargs['Feed']
         super().from_dict(**kwargs)
+        if hasattr(self, 'settings'):
+            self.settings = json.loads(self.settings)
 
 
 class MISPWarninglist(AbstractMISP):
@@ -1592,8 +1607,9 @@ class MISPEventDelegation(AbstractMISP):
 
 class MISPObjectAttribute(MISPAttribute):
 
-    _fields_for_feed: set = {'uuid', 'object_relation', 'value', 'category', 'type',
-                             'comment', 'data', 'timestamp', 'to_ids', 'disable_correlation'}
+    _fields_for_feed: set = {'uuid', 'value', 'category', 'type', 'comment', 'data',
+                             'deleted', 'timestamp', 'to_ids', 'disable_correlation',
+                             'first_seen', 'last_seen', 'object_relation'}
 
     def __init__(self, definition):
         super().__init__()
@@ -1640,7 +1656,7 @@ class MISPCommunity(AbstractMISP):
         super().from_dict(**kwargs)
 
     def __repr__(self):
-        return '<{self.__class__.__name__}(name={self.name}, uuid={self.uuid})'.format(self=self)
+        return f'<{self.__class__.__name__}(name={self.name}, uuid={self.uuid})'
 
 
 class MISPUserSetting(AbstractMISP):
@@ -1651,4 +1667,19 @@ class MISPUserSetting(AbstractMISP):
         super().from_dict(**kwargs)
 
     def __repr__(self):
-        return '<{self.__class__.__name__}(name={self.setting}'.format(self=self)
+        return f'<{self.__class__.__name__}(name={self.setting}'
+
+
+class MISPInbox(AbstractMISP):
+
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+        self.data: Dict
+
+    def from_dict(self, **kwargs):
+        if 'Inbox' in kwargs:
+            kwargs = kwargs['Inbox']
+        super().from_dict(**kwargs)
+
+    def __repr__(self):
+        return f'<{self.__class__.__name__}(name={self.type})>'
